#!/usr/bin/perl -w
########################################################
#
# This script creates jam files used to program TDCs,
# configuration file with parameters for TRBs etc...
# Run the script with "-h" option to get a help.
# Sergey Yurevich, 17.10.2007
#
########################################################

use strict;
use warnings;
use Data::Dumper;
use FileHandle;
use Getopt::Long;

#- the command line option flags
my $opt_help = 0; 
my $opt_jam  = 0;
my $opt_tcl  = 0;
my @opt_gen;

my $resultOpt = GetOptions ('h|help'  => \$opt_help,
			    'g|gen=i' => \@opt_gen,
			    'j|jam'   => \$opt_jam,
			    't|tcl'   => \$opt_tcl);

if(&checkOptions) {
    exit(0);
}

my $default_TDC_setup    = &readDefaultTDCsetup2hash();      #used by Marek
my $default_TDC_channels = &readDefaultTDCchannels2hash();   #used by Marek
my $default_TRB_setup    = &readDefaultTRBsetup2hash();      #used by Radek

#- read configuration file with changes for all TDCs on all TRBs
my $TRB_settings = &readNewTRBTDCsettings();

if($opt_jam) {
    #- check if all TDC settings on all TRBs are correct
    &checkTRBsettings();

    print "Write jam files with TRB TDC settings and channels ..... ";
    
    for my $TRB ( sort keys %$TRB_settings) {
	
	#- write TDC jam file with settings for every TRB 
	&writeTDCsettigns( $TRB );
	
	#- write TDC jam files with channels (on/off) for every TRB
	&writeTDCchannels( $TRB );
    }
}

if($opt_tcl) {
    print "Write files with TRB settings ..... ";
    
    #- write TRB tcl file with settings for all TRBs
    &writeTRBsettings();

    #- write TRB configuration file with settings for all TRBs
    &writeTRBConfSettings()
}

print "done\n";

exit(0);

sub getSortedArray()
{
    #- this subroutine fills the array (@bits) with direct 
    #- correspondence between single bit index and its value

    my ($default_TDC_setup_ref) = @_;

    my @bits = ();

    for my $var (keys %$default_TDC_setup_ref) {
	for my $index (keys %{$default_TDC_setup_ref->{ $var } }) {

	    my $value = $default_TDC_setup_ref->{ $var }->{ $index };
	    my @vbits = ();

	    if($index =~ /(\d+):(\d+)/) {
		my ($ind1, $ind2) = ($1, $2);

		#- split the bits of the value
		@vbits = split(//, $value);
		my $i = $#vbits; # index of manimal element

		foreach my $bit ($ind2..$ind1) {

		    #- fill in reverse order!!!
		    $bits[$bit] = $vbits[$i];
		    $i--;
		}
	    }
	    else {
		$bits[$index] = $value;
	    }
	}
    }

    return \@bits;
}

sub readDefaultTDCsetup2hash
{
    #- default configuration file for TDC
    my $setup = "./config/default_TDC_setup.conf";
    my $fh    = FileHandle->new("<$setup");
    if(!$fh) {
	print "Could not open \"$setup\". \n";
	exit;
    }

    my @file = <$fh>;
    $fh->close;

    my %TDCsetup = ();

    foreach my $line (@file) {
	chop($line);

	#- extract Variables, Indexes and Values
	if($line =~ /(\w+)\[(.*)\]=b(\d+).*/) {
	    my $var   = $1;
	    my $index = $2;
	    my $value = $3;

	    if($value =~ /[^01]/) {
		print "\nError: ($setup) $var\[$index\] has a wrong value: $value.\n";
		print "exit.";
		exit(0);
	    }

	    if(&checkBitSettings( $setup, "", $var, $index, $value)) {
		print "exit.";
		exit(0);
	    }

	    #$TDCsetup{ $var }{ 'index' } = $index;
	    #$TDCsetup{ $var }{ 'value' } = $value;
	    $TDCsetup{ $var }{ $index } = $value;
	}
	elsif($line =~ /(\w+)\[(.*)\]=b.*/) {
	    my $var   = $1;
	    my $index = $2;

	    print "\nError: ($setup) $var\[$index\] has a wrong value.\n";
	    print "exit.";
	    exit(0);
	}
    }

    return \%TDCsetup;
}

sub setValueDefaultTDCsetup()
{
    my ($var, $value, $hash_ref) = @_;

    for my $index ( keys %{$hash_ref->{ $var }}) {
 	$hash_ref->{ $var }->{ $index } = $value;
    }
}

sub readDefaultTDCchannels2hash
{
    #- default configuration file with channels on/off for TDC
    my $setup = "./config/default_TDC_channels.conf";
    my $fh    = FileHandle->new("<$setup");
    if(!$fh) {
	print "\nCould not open \"$setup\". \n";
	print "exit.\n";
	exit(0);
    }

    my @file = <$fh>;
    $fh->close;
    
    my %TDCchannels = ();
    
    foreach my $line (@file) {
	chop($line);

	#- extract Variables, Indexes and Values
	if($line =~ /(\w+)\[(.*)\]=b(\d+).*/) {
	    my $var   = $1;
	    my $index = $2;
	    my $value = $3;

	    if($value =~ /[^01]/) {
		print "\nError: ($setup) $var\[$index\] has a wrong value: $value\n";
		print "exit.";
		exit(0);
	    }
	    
	    if(&checkBitSettings( $setup, "", $var, $index, $value)) {
		print "exit.";
		exit(0);
	    }

	    $TDCchannels{ $index } = $value;
	}
	elsif($line =~ /(\w+)\[(.*)\]=b.*/) {
	    my $var   = $1;
	    my $index = $2;
	    
	    print "\nError: ($setup) $var\[$index\] has a wrong value.\n";
	    print "exit.";
	    exit(0);
	}
    }

    return \%TDCchannels;
}

sub readDefaultTRBsetup2hash
{
    #- read configuration file with settings for TRB 
    my $config_file = "./config/default_TRB_setup.conf";
    my $settings;
    unless ($settings = do $config_file) {
	die "Couldn't parse $config_file: $@, stopped"  if $@;
	die "Couldn't do $config_file: $!, stopped"     unless defined $settings;
	die "Couldn't run $config_file, stopped"        unless $settings;	
    }
    
    return $settings;
}

sub readNewTRBTDCsettings
{
    #- read configuration file with changes for all TDCs on all TRBs
    my $config_file = "./config/TRB_TDC_settings.conf";
    my $settings;
    unless ($settings = do $config_file) {
	die "Couldn't parse $config_file: $@, stopped"  if $@;
	die "Couldn't do $config_file: $!, stopped"     unless defined $settings;
	die "Couldn't run $config_file, stopped"        unless $settings;	
    }
    
    return $settings;
}

sub checkBitSettings()
{

    my ($TRB, $TDC, $var, $index, $value) = @_;

    my $error = 0;

    #- check consistency of Indexes and Values:
    #- number of bits in the Index should correspond
    #- to a number of bits in the value.
    if($index =~ /(\d+):(\d+)/) {
	my $diff = $1 - $2 + 1;

	$_ = $value; # needed for tr/0-1//
	unless(tr/0-1// == $diff) {
	    printf "\nError: ($TRB $TDC) $var\[$index\]=$value Value does not match number of bits!\n";
	    $error = 1;
	}
    }
    else {
	my $diff = 1;
	
	$_ = $value; # needed for tr/0-1//
	unless(tr/0-1// == $diff) {
	    printf "\nError: ($TRB $TDC) $var\[$index\]=$value Value does not match number of bits!\n";
	    $error = 1;
	}
    }

    return $error;
} 

sub writeTDCsettigns()
{
    my ($TRB) = @_;

    #- this subroutine writes TDC settings to a jam file
    #- for all four TDCs on a given TRB.

    my $outname = sprintf("./outdata/%s_TDC_setup.jam", $TRB); 
    my $fh_TDCsettings = FileHandle->new(">$outname");

    #- write the beginning of the file
    print $fh_TDCsettings &getTDCsettingsStart();

    for my $TDC (sort keys %{$TRB_settings->{$TRB}->{'TDC'}}) {

        #- get reference to a copy of a hash with default TDC setup
	my %default_TDC_setup_copy = %$default_TDC_setup;  # copy a hash
	my $default_TDC_setup_ref  = \%default_TDC_setup_copy; # get a reference of the hash copy

	my $trigger_count_offset = $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'trigger_count_offset'};
	my $matching_window      = $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'matching_window'};
	my $pulse_length         = $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'pulse_length'};
	my $reject_count_offset  = $trigger_count_offset + 50;
	my $search_window        = $matching_window + $pulse_length;
	
	my $reject_count_offset_str  = sprintf("%12.12b", 0xFFF - $reject_count_offset / 25);
	my $search_window_str        = sprintf("%12.12b", $search_window / 25);
	my $matching_window_str      = sprintf("%12.12b", $matching_window / 25);
	my $trigger_count_offset_str = sprintf("%12.12b", 0xFFF - $trigger_count_offset / 25);

	&setValueDefaultTDCsetup("match_window",         $matching_window_str,      $default_TDC_setup_ref);
	&setValueDefaultTDCsetup("search_window",        $search_window_str,        $default_TDC_setup_ref);
	&setValueDefaultTDCsetup("reject_count_offset",  $reject_count_offset_str,  $default_TDC_setup_ref);
	&setValueDefaultTDCsetup("trigger_count_offset", $trigger_count_offset_str, $default_TDC_setup_ref);

	if( $TDC eq "TDC_A") { &setValueDefaultTDCsetup("tdc_id", "0000", $default_TDC_setup_ref); }
	if( $TDC eq "TDC_B") { &setValueDefaultTDCsetup("tdc_id", "0001", $default_TDC_setup_ref); }
	if( $TDC eq "TDC_C") { &setValueDefaultTDCsetup("tdc_id", "0010", $default_TDC_setup_ref); }
	if( $TDC eq "TDC_D") { &setValueDefaultTDCsetup("tdc_id", "0011", $default_TDC_setup_ref); }

	my $totalValue = "";
	my $indexMax = "";   # needed for a check of coincedense
	
	#- check all settings.
	for my $var (keys %$default_TDC_setup_ref) {
	    for my $index (keys %{$default_TDC_setup_ref->{ $var }}) {
		
		my $value = $default_TDC_setup_ref->{$var}->{ $index };

		if(&checkBitSettings( $TRB, $TDC, $var, $index, $value)) {
		    $fh_TDCsettings->close();
		    print "exit.";
		    exit(0);
		}
	    }
	}
	
	#- fill array with direct correspondence: single bit index => value
	my $bits = &getSortedArray( $default_TDC_setup_ref );

	#- join all values to a total value
	foreach my $value (@$bits) {
	    $totalValue = "$totalValue" . "$value";
	}
	
	#- get a size of the @bits
	my @bits = @$bits; # dereference
	$indexMax = $#bits+1;

	#- check coincedense between maximal bit number 
	#- and the length of the total value
	my $lengthTotValue = length($totalValue);

	unless($indexMax == length($totalValue)) {	
	    print "\nError: ($TRB, $TDC) total number of bits does not match the total value for TDC settings.\n";
	    $fh_TDCsettings->close();
	    print "exit.\n";
	    exit(0);
	}
	
	#- get parity of the total value
	my $parity = &getParity( $totalValue );
	
	#- add parity to the total value
	$totalValue = "$totalValue" . "$parity";

	my $line = sprintf("BOOLEAN WRITE_DATA_%s[647]= BIN %s%s\n", $TDC, $totalValue, ";"); 

	#- write settings for a given TDC into the file
	print $fh_TDCsettings $line;
    }

    #- write the end of the file
    print $fh_TDCsettings &getTDCsettingsEnd();
    $fh_TDCsettings->close();
}    

sub writeTDCchannels()
{
    my ($TRB) = @_;

    #- this subroutine writes TDC channels (on/off) to a jam file
    #- for all four TDCs on a given TRB.

    my $outname = sprintf("./outdata/%s_TDC_channels.jam", $TRB); 
    my $fh_TDCchannels = FileHandle->new(">$outname");    

    #- write the beginning of the file
    print $fh_TDCchannels &getTDCchannelsStart();

    #- reference to a hash with additional bits and characters 
    #- in the beginning and in the end of the BIN line
    my $TDCchannelsAddon = &getTDCchannelsAddon();

    #- loop over TDCs
    for my $TDC (sort keys %{$TRB_settings->{$TRB}->{'TDC'}}) {

	#- get reference to a copy of a hash with default TDC channels
	my %default_TDC_channels_copy = %$default_TDC_channels;  # copy a hash
	my $default_TDC_channels_ref  = \%default_TDC_channels_copy; # get a reference of the hash copy

	#- loop over indexes of the channels for a given TDC
	for my $index (sort keys %{$TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'enable_channel'}}) {
	
	    my $value = $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'enable_channel'}->{$index};
	    if($index =~ /(\d+):(\d+)/) {
		foreach my $bit ($1..$2) {
		    $default_TDC_channels_ref->{ $bit } = $value;
		}
	    }
	    else {
		$default_TDC_channels_ref->{ $index } = $value;
	    }
	}
	
	my $totalValue = "";
	my $indexMax = "";   # needed for a check of coincedense
	
	#- check all channels
	for my $index ( sort { $a <=> $b } keys %$default_TDC_channels_ref) { # numerical sort
	    my $value = $default_TDC_channels_ref->{$index};
	    
	    if(&checkBitSettings( $TRB, $TDC, "enable_channel", $index, $value)) {
		$fh_TDCchannels->close();
		print "exit.";
		exit(0);
	    }
	    
	    $totalValue = "$totalValue" . "$value";
	    $indexMax   = $index;
	}

	#- check coincedense between maximal bit number 
	#- and the length of the total value 
	unless($indexMax == length($totalValue) - 1) {	
	    print "\nError: ($TRB, $TDC) total number of bits does not match the total value for TDC channels.\n";
	    $fh_TDCchannels->close();
	    print "exit.\n";
	    exit(0);
	}

	#- the channel settings will be written 6 times into a file
	#- every time with different addons accordning to $TDCchannelsAddon
	for my $data (sort keys %$TDCchannelsAddon) {

	    my $totalValueAdd = "$TDCchannelsAddon->{$data}->{ 'startbits' }" . 
				"$totalValue" .
				"$TDCchannelsAddon->{$data}->{ 'endbits' }";

	    #- get parity of the total value
	    my $parity = &getParity($totalValueAdd);
	    
	    my $addstartline = sprintf("%s_%s%s",$TDCchannelsAddon->{$data}->{ 'startline' },
				                 $TDC,"[40]= BIN ");

	    $totalValueAdd = "$addstartline" . "$totalValueAdd" . "$parity" . 
		"$TDCchannelsAddon->{$data}->{ 'endline' }\n";
	    
	    #- write channels for a given TDC into the file
	    print $fh_TDCchannels $totalValueAdd;
	}	
    }

    #- write the end of the file
    print $fh_TDCchannels &getTDCchannelsEnd();
    $fh_TDCchannels->close();
}

sub writeTRBsettings()
{
    #- this subroutine writes TRB settings to a tcl file for all TRBs.

    my $outname = sprintf("./outdata/trb_p.tcl");
    my $fh_TRBsettings = FileHandle->new(">$outname");
    
    #- loop over all TRBs
    for my $TRB ( sort keys %$TRB_settings) {
	
	#- get reference to a copy of a hash with default TRB setup
	my %default_TRB_setup_copy = %$default_TRB_setup;  # copy a hash
	my $default_TRB_setup_ref  = \%default_TRB_setup_copy; # get a reference of the hash's copy
	
	#- copy all new settings to a default hash copy
	for my $par (sort keys %{$TRB_settings->{$TRB}->{'TRB'}}) {

	    $default_TRB_setup_ref->{ $par } = $TRB_settings->{$TRB}->{'TRB'}->{$par}; 
	}
	
	#- increment sub_evt_id according to TRB number
	unless( $TRB_settings->{$TRB}->{'TRB'}->{'sub_evt_id'} ) {
	    if( $TRB =~ /TRB_(\d+)/) {

		my $trbnum = 1*$1 - 1; #1 is subtracted since trb1 must not get an increment
		my $dec_subevtid = hex($default_TRB_setup->{'sub_evt_id'});
		my $hex_subevtid    = sprintf("0x%x", $dec_subevtid+$trbnum);
		$default_TRB_setup_ref->{ 'sub_evt_id' } = $hex_subevtid; 
	    }
	}

	#- write settings for a given TRB
	for my $par (sort keys %{$default_TRB_setup_ref}) {
	    if( $TRB =~ /TRB_(\d+)/) {
		my $trbnum = 1*$1;
		my $parameter = "set " . "trb" . "$trbnum" . "(" . "$par" . ") " . 
		                "$default_TRB_setup_ref->{ $par }\n";
	    
		print $fh_TRBsettings $parameter;
	    }
	}
    }
    
    $fh_TRBsettings->close();
}

sub writeTRBConfSettings()
{
    #- this subroutine writes (perl-like) configuration file with TRB parameters.

    my $outname = sprintf("./outdata/trb.conf");
    my $fh_TRBsettings = FileHandle->new(">$outname");

    print $fh_TRBsettings "# Do not make changes here!\n";
    print $fh_TRBsettings "# This file is automatically generated.\n\n";
    print $fh_TRBsettings "# Contains parameters to configure TRBs.\n\n";
    print $fh_TRBsettings "my %TRB_setup = (\n";
        
    #- loop over all TRBs
    for my $TRB ( sort keys %$TRB_settings) {
	
	print $fh_TRBsettings "'" . "$TRB" . "' => {\n";

	#- get reference to a copy of a hash with default TRB setup
	my %default_TRB_setup_copy = %$default_TRB_setup;  # copy a hash
	my $default_TRB_setup_ref  = \%default_TRB_setup_copy; # get a reference of the hash's copy
	
	#- copy all new settings to a default hash copy
	for my $par (sort keys %{$TRB_settings->{$TRB}->{'TRB'}}) {

	    $default_TRB_setup_ref->{ $par } = $TRB_settings->{$TRB}->{'TRB'}->{$par}; 
	}
	
	#- increment sub_evt_id according to TRB number if not already defined
	unless( $TRB_settings->{$TRB}->{'TRB'}->{'sub_evt_id'} ) {
	    if( $TRB =~ /TRB_(\d+)/) {

		my $trbnum = 1*$1 - 1; #1 is subtracted since trb1 must not get an increment
		my $dec_subevtid = hex($default_TRB_setup->{'sub_evt_id'});
		my $hex_subevtid    = sprintf("0x%x", $dec_subevtid+$trbnum);
		$default_TRB_setup_ref->{ 'sub_evt_id' } = $hex_subevtid; 
	    }
	}

	#- write settings for a given TRB
	for my $par (sort keys %{$default_TRB_setup_ref}) {

	    my $parameter;
	    if( $par eq "wmark" ){
		$parameter = "         '" . "$par" ."' => " . 
		    "$default_TRB_setup_ref->{ $par }" . ",\n";
	    }
	    else{
		$parameter = "         '" . "$par" ."' => '" . 
		    "$default_TRB_setup_ref->{ $par }" . "',\n";
	    }

	    print $fh_TRBsettings $parameter;
	}

	print $fh_TRBsettings "},\n";
    }
    
    print $fh_TRBsettings ");\n\n";
    print $fh_TRBsettings "return \\%TRB_setup;\n";
    
    $fh_TRBsettings->close();
}

sub getParity()
{
    my ($totValue) = @_;

    #- count how many ones we have in a string
    my $count = ($totValue =~ tr/1//);

    my $parity = 1;

    if( $count % 2 == 0) {
	$parity = 0;
    }

    return $parity;
}

sub checkTRBsettings
{
    printf "Checking TRB TDC settings .......";

    for my $TRB ( sort keys %$TRB_settings) {
	for my $TDC (sort keys %{$TRB_settings->{$TRB}->{'TDC'}}) {

	    if( $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'matching_window'} % 25 != 0) {
		printf "\nError: ($TRB $TDC) matching_window has to be given in multiples of 25 ns\n";
		printf "exit.\n";
		exit(0);
	    }
	    if( $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'trigger_count_offset'} % 25 != 0) {
		printf "\nError: ($TRB $TDC) trigger_count_offset has to be given in multiples of 25 ns\n";
		printf "exit.\n";
		exit(0);
	    }
	    if( $TRB_settings->{$TRB}->{'TDC'}->{$TDC}->{'pulse_length'} % 25 != 0) {
		printf "\nError: ($TRB $TDC) pulse_length has to be given in multiples of 25 ns\n";
		printf "exit.\n";
		exit(0);
	    }
	}
    }
    
    printf " done\n";
}

sub getTDCchannelsAddon
{
    #- hash with additional bits to create a proper jam file
    my %TDCchannelsAddon = (
	      'datas1' => {
		  'startline' => "BOOLEAN WRITE_DATAS1",
		  'startbits' => "00100",
		  'endbits'   => "11",
		  'endline'   => ";'PLL=1;DLL=1,pattern=0100",
	      },
	      'datas2' => {
		  'startline' => "BOOLEAN WRITE_DATAS2",
		  'startbits' => "00100",
		  'endbits'   => "10",
		  'endline'   => ";'PLL=0;DLL=1,pattern=0100",
	      },
	      'datas3' => {
		  'startline' => "BOOLEAN WRITE_DATAS3",
		  'startbits' => "00100",
		  'endbits'   => "00",
		  'endline'   => ";'PLL=0;DLL=0,pattern=0100",
	      },	  
	      'datas4' => {
		  'startline' => "BOOLEAN WRITE_DATAS4",
		  'startbits' => "00101",
		  'endbits'   => "00",
		  'endline'   => ";'PLL=0;DLL=0,pattern=0100,Global reset=1",
	      },
	      'datas5' => {
		  'startline' => "BOOLEAN WRITE_DATAS5",
		  'startbits' => "00100",
		  'endbits'   => "00",
		  'endline'   => ";'PLL=0;DLL=0,pattern=0100,Global reset=0",
	      },
	      'datas6' => {
		  'startline' => "BOOLEAN WRITE_DATAS6",
		  'startbits' => "10100",
		  'endbits'   => "00",
		  'endline'   => ";'PLL=0;DLL=0,pattern=0101",
	      },
    );

    return \%TDCchannelsAddon;
}

sub getTDCsettingsStart
{
    my $TDCstart = <<EOF;
BOOLEAN ID_CODE5[5]= BIN 00011;
BOOLEAN BYPASS5[5]= BIN 11110;
BOOLEAN BYPASS10[10]= BIN 1111011110;
BOOLEAN BYPASS15[15]= BIN 111101111011110;
BOOLEAN CAPTCOMP;
BOOLEAN CAPTARAY[647];
EOF
    return $TDCstart;
}

sub getTDCsettingsEnd
{
    my $TDCend = <<EOF;
INTEGER i;
FOR i=0 to 646;
LET CAPTARAY[i]=ID_CODE5[4];
NEXT i;
PREDR 3;
PREIR 15, BYPASS15[0..14];
POSTDR 0;
POSTIR 0;
IRSCAN 5, ID_CODE5[0..4];
DRSCAN 647, WRITE_DATA_TDC_A[0..646];
DRSCAN 647, WRITE_DATA_TDC_A[0..646], COMPARE WRITE_DATA_TDC_A[0..646],CAPTARAY[0..646],CAPTCOMP ;
IF CAPTCOMP==0 THEN PRINT "Setup data error!";
IF CAPTCOMP==1 THEN PRINT "Setup data passed";
PREDR 2;
PREIR 10 , BYPASS10[0..9];
POSTDR 1;
POSTIR 5, BYPASS5[0..4];
IRSCAN 5, ID_CODE5[0..4];
DRSCAN 647, WRITE_DATA_TDC_B[0..646];
DRSCAN 647, WRITE_DATA_TDC_B[0..646], COMPARE WRITE_DATA_TDC_B[0..646],CAPTARAY[0..646],CAPTCOMP ;
IF CAPTCOMP==0 THEN PRINT "Setup data error!";
IF CAPTCOMP==1 THEN PRINT "Setup data passed";
PREDR 1;
PREIR 5 ,BYPASS5[0..4];
POSTDR 2;
POSTIR 10,BYPASS10[0..9];
IRSCAN 5, ID_CODE5[0..4];
DRSCAN 647, WRITE_DATA_TDC_C[0..646];
DRSCAN 647, WRITE_DATA_TDC_C[0..646], COMPARE WRITE_DATA_TDC_C[0..646],CAPTARAY[0..646],CAPTCOMP ;
IF CAPTCOMP==0 THEN PRINT "Setup data error!";
IF CAPTCOMP==1 THEN PRINT "Setup data passed";
PREDR 0;
PREIR 0 ;
POSTDR 3;
POSTIR 15, BYPASS15[0..14];
IRSCAN 5, ID_CODE5[0..4];
DRSCAN 647, WRITE_DATA_TDC_D[0..646];
DRSCAN 647, WRITE_DATA_TDC_D[0..646], COMPARE WRITE_DATA_TDC_D[0..646],CAPTARAY[0..646],CAPTCOMP ;
IF CAPTCOMP==0 THEN PRINT "Setup data error!";
IF CAPTCOMP==1 THEN PRINT "Setup data passed";
STATE IDLE;
EXIT 0;
EOF


    return $TDCend
}

sub getTDCchannelsEnd
{
    my $TDCend = <<EOF;
WAIT 3000 CYCLES, 300000 usec;
INTEGER i;
FOR i=0 to 39;
LET CAPTARAY[i]=ID_CODE1[3];
NEXT i;
STATE RESET;
IRSTOP IRPAUSE;
STATE IDLE;
PREDR 0;
PREIR 0 ;
POSTDR 3;
POSTIR 15,BYPASS15[0..14];
IRSCAN 5, ID_CODE1[0..4];
DRSCAN 40, WRITE_DATAS1_TDC_D[0..39];
WAIT 3000 CYCLES, 300000 usec;
DRSCAN 40, WRITE_DATAS2_TDC_D[0..39];
WAIT 3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS3_TDC_D[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS4_TDC_D[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS5_TDC_D[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS6_TDC_D[0..39];
PREDR 1;
PREIR 5,BYPASS5[0..4] ;
POSTDR 2;
POSTIR 10,BYPASS10[0..9];
IRSCAN 5, ID_CODE1[0..4];
DRSCAN 40, WRITE_DATAS1_TDC_C[0..39];
WAIT  3000 CYCLES,30000 usec;
DRSCAN 40, WRITE_DATAS2_TDC_C[0..39];
WAIT  3000 CYCLES,30000 usec;
DRSCAN 40, WRITE_DATAS3_TDC_C[0..39];
WAIT  3000 CYCLES,30000 usec;
DRSCAN 40, WRITE_DATAS4_TDC_C[0..39];
WAIT  3000 CYCLES,30000 usec;
DRSCAN 40, WRITE_DATAS5_TDC_C[0..39];
WAIT  3000 CYCLES,30000 usec;
DRSCAN 40, WRITE_DATAS6_TDC_C[0..39];
PREDR 2;
PREIR 10 , BYPASS10[0..9];
POSTDR 1;
POSTIR 5, BYPASS5[0..4];
IRSCAN 5, ID_CODE1[0..4];
DRSCAN 40, WRITE_DATAS1_TDC_B[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS2_TDC_B[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS3_TDC_B[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS4_TDC_B[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS5_TDC_B[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS6_TDC_B[0..39];
PREDR 3;
PREIR 15,BYPASS15[0..14];
POSTDR 0;
POSTIR 0;
IRSCAN 5, ID_CODE1[0..4];
DRSCAN 40, WRITE_DATAS1_TDC_A[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS2_TDC_A[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS3_TDC_A[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS4_TDC_A[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS5_TDC_A[0..39];
WAIT  3000 CYCLES,300000 usec;
DRSCAN 40, WRITE_DATAS6_TDC_A[0..39];
STATE IDLE;
EXIT 0;
EOF
    return $TDCend;
}

sub getTDCchannelsStart
{
    my $TDCstart = <<EOF;
BOOLEAN ID_CODE1[5]= BIN 10010;\'Control register
BOOLEAN BYPASS5[5]= BIN 11110;
BOOLEAN BYPASS10[10]= BIN 1111011110;
BOOLEAN BYPASS15[15]= BIN 111101111011110;
BOOLEAN CAPTCOMP;
BOOLEAN CAPTARAY[40];
EOF

    return $TDCstart;
}

sub genSampleTRBTDCsettigns()
{
    my ($genstart, $genend) = @_;
    
    # this subroutine generates sample TRB_TDC_settings.conf 
    # for a given number of TRBs ($opt_gen)

    my $outname = "./config/TRB_TDC_settings.conf.sample"; 
    my $fh_TRBsample = FileHandle->new(">$outname"); 

    my $TRBsampleStart = <<EOF;
# This is a configuration file with NEW settings for TRBs and TDCs.
# It provides special settings for TRBs, channels on/off and windows for TDCs.
# 'enable_channel' => index of channel (0,2,..,4:9,..) => 1/0 (on/off)
# This file is read by make.pl
#
# All TDC values are in ns, granularity is 25ns
# trigger_count_offset : time from digital trigger (+250ns) to opening of
#                           window. This window has to be larger than
#                           matching_window + 50ns + pulse_length
# matching_window :      size of window
# pulse_length :         max. length of an input pulse

my %TRB_TDC_setup = (
EOF

    print $fh_TRBsample $TRBsampleStart;

    #- loop over all TRBs
    foreach my $TRBnum ($genstart..$genend) {

	my $TRB = sprintf("TRB_%03d",$TRBnum);

	print $fh_TRBsample "'$TRB' => {\n";

	#- print sample settings for TRB (only for first TRB)
	if( $TRBnum == 1 ) {
	    print $fh_TRBsample "        'TRB' => {\n";
	    print $fh_TRBsample "                'wmark' => 32768,\n";
	    print $fh_TRBsample "	 },\n";
	}

	#- print sample settings for TDCs
	print $fh_TRBsample "        'TDC' => {\n";
	
	#- loop over four TDCs
	foreach my $TDCnum ('A'..'D') {
	    
	    my $TDC = sprintf("TDC_%s",$TDCnum);
	    
	    print $fh_TRBsample "                '$TDC' => {\n";
	    print $fh_TRBsample "                        'trigger_count_offset'   => 1200,\n";
	    print $fh_TRBsample "                        'matching_window'        => 1175,\n";
	    print $fh_TRBsample "                        'pulse_length'           => 25,\n";

	    #- print enable_channel only for first TDC on first TRB 
	    if($TRBnum == 1 && $TDCnum eq "A") {
		print $fh_TRBsample "                        'enable_channel'         => {\n";
		print $fh_TRBsample "                                                        '0' => 1,\n";
		print $fh_TRBsample "                                                        '2' => 1,\n";
		print $fh_TRBsample "                                                        '4:9' => 1,\n";
		print $fh_TRBsample "                        },\n";
	    }
	    
	    print $fh_TRBsample "                },\n";
	}
	
	print $fh_TRBsample "        },\n";
	print $fh_TRBsample "},\n";
    }

my $TRBsampleEnd = <<EOF;
);

return \\%TRB_TDC_setup;
EOF

    print $fh_TRBsample $TRBsampleEnd;
}

sub usage()
{
   print << 'EOF';
Usage: make.pl [-h|--help] [-g|--gen TRB_first_num -g|--gen TRB_last_num]
EOF
}

sub help()
{
    print << 'EOF';

This script creates jam files with settings for each TRB for all four TDCs:
outdata/TRB_001_TDC_setup.jam     
outdata/TRB_001_TDC_channels.jam 

the script creates also tcl file with settings for all TRBs:
outdata/trb_p.tcl
and configuration file with TRB settings for all TRBs:
outdata/trb.conf    (this file is read by the start_daq script)

It requires the following input information:
config/default_TDC_setup.conf     = default settings for TDC
config/default_TDC_channels.conf  = default settings for TDC 
                                    channels (on/off)
config/default_TRB_setup.conf     = default settings for TRB
config/TRB_TDC_settings.conf      = new settings which should 
                                    overwrite default settings

Example of TRB_TDC_settings.conf for one TDC:

'TRB_001' => {
        'TDC' => {
                'TDC_A' => {
                        'trigger_count_offset'   => 1200,
                        'matching_window'        => 1175,
                        'pulse_length'           => 25,
                        'enable_channel'         => {
                                                        '0' => 0,
                                                        '2' => 0,
                                                        '4:9' => 0,
                        },
		},

Command line: make.pl [-h|--help] [-g|--gen TRB_first_num -g|--gen TRB_last_num] [-j|--jam] [-t|--tcl]

-h : print this help.
-g : generate sample TRB_TDC_settings.conf for a given number of TRBs.
     First TRB will have number TRB_first_num and last TRB will have 
     number TRB_last_num. The settings will be written into the sample file:
     config/TRB_TDC_settings.conf.sample (do not forget to rename it to 
     config/TRB_TDC_settings.conf).
-j : the script reads configuration files from config/. and writes 
     out jam files to outdata/.
-t : the script reads configuration file from config/. and writes 
     out tcl file and trb.conf file to outdata/. 

EOF
}

sub checkOptions
{
    if(!$resultOpt ) {
	&usage();
	return 1;
    }
    
    if($opt_help) {
	&help();
	return 1;
    }
    
    if(@opt_gen) {
	if(@opt_gen != 2) {
	    print "Usage: make.pl [-g|--gen TRB_first_num -g|--gen TRB_last_num]\n\n";
	    return 1;
	}
	
	my ( $genstart, $genend) = @opt_gen;
	&genSampleTRBTDCsettigns( $genstart, $genend);
	return 1;
    }
    
    if(!$opt_jam && !$opt_tcl) {
	print "Usage: make.pl [-j|--jam] [-t|--tcl]\n";
	print "Get help: make.pl -h\nn";
	return 1;
    }

    return 0;
}



